@@ -1,7 +1,7 @@
 #ifndef AUTH_COMMON_H
 # define AUTH_COMMON_H
 
-#define MAX_TRIES 3
+#define MAX_PASSWORD_TRIES 3
 
 extern const char* pass_msg_first;
 extern const char* pass_msg_second;
