@@ -27,7 +27,6 @@
 struct radius_ctx_st {
 	char username[MAX_USERNAME_SIZE*2];
 	char groupname[MAX_GROUPNAME_SIZE];
-	char sid[BASE64_LENGTH(SID_SIZE) + 1];
 
 	char remote_ip[MAX_IP_STR];
 
@@ -46,7 +45,6 @@ struct radius_ctx_st {
 	char **routes;
 	unsigned routes_size;
 
-	const char *config;	/* radius config file */
 	const char *pass_msg;
 	unsigned retries;
 };
