@@ -33,21 +33,30 @@
 #include <c-ctype.h>
 #include <auth/pam.h>
 #include <auth/radius.h>
+#include <acct/pam.h>
+#include <acct/radius.h>
 #include <auth/plain.h>
+#include <auth/gssapi.h>
+#include <auth/common.h>
 #include <sec-mod-sup-config.h>
+#include <sec-mod-acct.h>
+
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <netdb.h>
 
 #include <vpn.h>
 #include <cookies.h>
 #include <main.h>
 #include <ctl.h>
 #include <tlslib.h>
+#include "cfg.h"
 
 #define OLD_DEFAULT_CFG_FILE "/etc/ocserv.conf"
 #define DEFAULT_CFG_FILE "/etc/ocserv/ocserv.conf"
 
 static char pid_file[_POSIX_PATH_MAX] = "";
 static const char* cfg_file = DEFAULT_CFG_FILE;
-static const struct auth_mod_st *amod = NULL;
 
 struct cfg_options {
 	const char* name;
@@ -58,16 +67,22 @@ struct cfg_options {
 
 static struct cfg_options available_options[] = {
 	{ .name = "auth", .type = OPTION_MULTI_LINE, .mandatory = 1 },
+	{ .name = "enable-auth", .type = OPTION_MULTI_LINE, .mandatory = 0 },
 	{ .name = "route", .type = OPTION_MULTI_LINE, .mandatory = 0 },
 	{ .name = "no-route", .type = OPTION_MULTI_LINE, .mandatory = 0 },
 	{ .name = "select-group", .type = OPTION_MULTI_LINE, .mandatory = 0 },
 	{ .name = "custom-header", .type = OPTION_MULTI_LINE, .mandatory = 0 },
 	{ .name = "split-dns", .type = OPTION_MULTI_LINE, .mandatory = 0 },
+	{ .name = "acct", .type = OPTION_STRING, .mandatory = 0 },
 	{ .name = "listen-host", .type = OPTION_STRING, .mandatory = 0 },
 	{ .name = "listen-host-is-dyndns", .type = OPTION_BOOLEAN, .mandatory = 0 },
 	{ .name = "compression", .type = OPTION_BOOLEAN, .mandatory = 0 },
 	{ .name = "no-compress-limit", .type = OPTION_NUMERIC, .mandatory = 0 },
 	{ .name = "tcp-port", .type = OPTION_NUMERIC, .mandatory = 0 },
+	{ .name = "max-ban-score", .type = OPTION_NUMERIC, .mandatory = 0 },
+	{ .name = "ban-points-wrong-password", .type = OPTION_NUMERIC, .mandatory = 0 },
+	{ .name = "ban-connection", .type = OPTION_NUMERIC, .mandatory = 0 },
+	{ .name = "ban-points-kkdcp", .type = OPTION_NUMERIC, .mandatory = 0 },
 	{ .name = "udp-port", .type = OPTION_NUMERIC, .mandatory = 0 },
 	{ .name = "keepalive", .type = OPTION_NUMERIC, .mandatory = 0 },
 	{ .name = "dpd", .type = OPTION_NUMERIC, .mandatory = 0 },
@@ -88,6 +103,9 @@ static struct cfg_options available_options[] = {
 	{ .name = "connect-script", .type = OPTION_STRING, .mandatory = 0 },
 	{ .name = "disconnect-script", .type = OPTION_STRING, .mandatory = 0 },
 	{ .name = "pid-file", .type = OPTION_STRING, .mandatory = 0 },
+#ifdef HAVE_GSSAPI
+	{ .name = "kkdcp", .type = OPTION_STRING, .mandatory = 0 },
+#endif
 	{ .name = "socket-file", .type = OPTION_STRING, .mandatory = 1 },
 	{ .name = "listen-clear-file", .type = OPTION_STRING, .mandatory = 0 },
 	{ .name = "occtl-socket-file", .type = OPTION_STRING, .mandatory = 0 },
@@ -121,6 +139,7 @@ static struct cfg_options available_options[] = {
 	{ .name = "mobile-idle-timeout", .type = OPTION_NUMERIC, .mandatory = 0 },
 	{ .name = "max-clients", .type = OPTION_NUMERIC, .mandatory = 0 },
 	{ .name = "min-reauth-time", .type = OPTION_NUMERIC, .mandatory = 0 },
+	{ .name = "ban-reset-time", .type = OPTION_NUMERIC, .mandatory = 0 },
 	{ .name = "max-same-clients", .type = OPTION_NUMERIC, .mandatory = 0 },
 
 	{ .name = "rx-data-per-sec", .type = OPTION_NUMERIC, .mandatory = 0 },
@@ -151,45 +170,6 @@ static struct cfg_options available_options[] = {
 	{ .name = "default-group-config", .type = OPTION_STRING, .mandatory = 0 },
 };
 
-#define get_brackets_string get_brackets_string1
-static char *get_brackets_string1(void *pool, const char *str);
-
-#ifdef HAVE_RADIUS
-static char *get_brackets_string2(void *pool, const char *str)
-{
-	char *p, *p2;
-	unsigned len;
-
-	p = strchr(str, '[');
-	if (p == NULL) {
-		return NULL;
-	}
-	p++;
-
-	p = strchr(p, ',');
-	if (p == NULL) {
-		return NULL;
-	}
-	p++;
-
-	while (c_isspace(*p))
-		p++;
-
-	p2 = strchr(p, ',');
-	if (p2 == NULL) {
-		p2 = strchr(p, ']');
-		if (p2 == NULL) {
-			fprintf(stderr, "error parsing %s\n", str);
-			exit(1);
-		}
-	}
-
-	len = p2 - p;
-
-	return talloc_strndup(pool, p, len);
-}
-#endif
-
 static const tOptionValue* get_option(const char* name, unsigned * mand)
 {
 unsigned j;
@@ -217,9 +197,9 @@ unsigned j;
 			} \
 		} \
 		do { \
-		        if (val && !strcmp(val->pzName, name)==0) \
+		        if (val && strcmp(val->pzName, name)!=0) \
 				continue; \
-		        s_name[num] = talloc_strdup(config, val->v.strVal); \
+		        s_name[num] = talloc_strdup(s_name, val->v.strVal); \
 		        num++; \
 		        if (num>=MAX_CONFIG_ENTRIES) \
 		        break; \
@@ -244,11 +224,11 @@ unsigned j;
 		} \
 		do { \
 		        char *xp; \
-		        if (val && !strcmp(val->pzName, name)==0) \
+		        if (val && strcmp(val->pzName, name)!=0) \
 				continue; \
 		        s_name[num] = talloc_strdup(config, val->v.strVal); \
 		        xp = strchr(s_name[num], '['); if (xp != NULL) *xp = 0; \
-		        s_name2[num] = get_brackets_string(config, val->v.strVal); \
+		        s_name2[num] = get_brackets_string1(config, val->v.strVal); \
 		        num++; \
 		        if (num>=MAX_CONFIG_ENTRIES) \
 		        break; \
@@ -260,15 +240,18 @@ unsigned j;
 		exit(1); \
 	}}
 
-#define READ_STRING(name, s_name) { \
+#define PREAD_STRING(pool, name, s_name) { \
 	val = get_option(name, &mand); \
 	if (val != NULL && val->valType == OPARG_TYPE_STRING) \
-		s_name = talloc_strdup(config, val->v.strVal); \
+		s_name = talloc_strdup(pool, val->v.strVal); \
 	else if (mand != 0) { \
 		fprintf(stderr, "Configuration option %s is mandatory.\n", name); \
 		exit(1); \
 	}}
 
+#define READ_STRING(name, s_name) \
+	PREAD_STRING(config, name, s_name)
+
 #define READ_STATIC_STRING(name, s_name) { \
 	val = get_option(name, &mand); \
 	if (val != NULL && val->valType == OPARG_TYPE_STRING) \
@@ -345,33 +328,6 @@ unsigned j;
 	}
 }
 
-static char *get_brackets_string1(void *pool, const char *str)
-{
-	char *p, *p2;
-	unsigned len;
-
-	p = strchr(str, '[');
-	if (p == NULL) {
-		return NULL;
-	}
-	p++;
-	while (c_isspace(*p))
-		p++;
-
-	p2 = strchr(p, ',');
-	if (p2 == NULL) {
-		p2 = strchr(p, ']');
-		if (p2 == NULL) {
-			fprintf(stderr, "error parsing %s\n", str);
-			exit(1);
-		}
-	}
-
-	len = p2 - p;
-
-	return talloc_strndup(pool, p, len);
-}
-
 
 /* Parses the string ::1/prefix, to return prefix
  * and modify the string to contain the network only.
@@ -395,12 +351,263 @@ unsigned extract_prefix(char *network)
 	return prefix;
 }
 
-const struct auth_mod_st *get_auth_mod(void)
+typedef struct auth_types_st {
+	const char *name;
+	unsigned name_size;
+	const struct auth_mod_st *mod;
+	unsigned type;
+	void *(*get_brackets_string)(struct perm_cfg_st *config, const char *);
+} auth_types_st;
+
+#define NAME(x) (x),(sizeof(x)-1)
+static auth_types_st avail_auth_types[] =
 {
-	return amod;
+#ifdef HAVE_PAM
+	{NAME("pam"), &pam_auth_funcs, AUTH_TYPE_PAM, pam_get_brackets_string},
+#endif
+#ifdef HAVE_GSSAPI
+	{NAME("gssapi"), &gssapi_auth_funcs, AUTH_TYPE_GSSAPI, gssapi_get_brackets_string},
+#endif
+#ifdef HAVE_RADIUS
+	{NAME("radius"), &radius_auth_funcs, AUTH_TYPE_RADIUS, radius_get_brackets_string},
+#endif
+	{NAME("plain"), &plain_auth_funcs, AUTH_TYPE_PLAIN, plain_get_brackets_string},
+	{NAME("certificate"), NULL, AUTH_TYPE_CERTIFICATE, NULL},
+};
+
+static void figure_auth_funcs(struct perm_cfg_st *config, char **auth, unsigned auth_size,
+			      unsigned primary)
+{
+	unsigned j, i;
+	unsigned found;
+
+	if (primary != 0) {
+		/* Set the primary authentication methods */
+		for (j=0;j<auth_size;j++) {
+			found = 0;
+			for (i=0;i<sizeof(avail_auth_types)/sizeof(avail_auth_types[0]);i++) {
+				if (c_strncasecmp(auth[j], avail_auth_types[i].name, avail_auth_types[i].name_size) == 0) {
+					if (avail_auth_types[i].get_brackets_string)
+						config->auth[0].additional = avail_auth_types[i].get_brackets_string(config, auth[j]+avail_auth_types[i].name_size);
+
+					if (config->auth[0].amod != NULL && avail_auth_types[i].mod != NULL) {
+						fprintf(stderr, "%s: You cannot mix multiple authentication methods of this type\n", auth[j]);
+						exit(1);
+					}
+
+					if (config->auth[0].amod == NULL)
+						config->auth[0].amod = avail_auth_types[i].mod;
+					config->auth[0].type |= avail_auth_types[i].type;
+					if (config->auth[0].name == NULL) {
+						config->auth[0].name = talloc_strdup(config, avail_auth_types[i].name);
+					} else {
+						char *tmp;
+						tmp = talloc_asprintf(config, "%s+%s", config->auth[0].name, avail_auth_types[i].name);
+						talloc_free(config->auth[0].name);
+						config->auth[0].name = tmp;
+					}
+					config->auth[0].enabled = 1;
+					config->auth_methods = 1;
+					found = 1;
+					break;
+				}
+			}
+
+			if (found == 0) {
+				fprintf(stderr, "Unknown or unsupported auth method: %s\n", auth[j]);
+				exit(1);
+			}
+			talloc_free(auth[j]);
+		}
+		fprintf(stderr, "Setting '%s' as primary authentication method\n", config->auth[0].name);
+	} else {
+		unsigned x = config->auth_methods;
+		/* Append authentication methods (alternative options) */
+		for (j=0;j<auth_size;j++) {
+			found = 0;
+			for (i=0;i<sizeof(avail_auth_types)/sizeof(avail_auth_types[0]);i++) {
+				if (c_strncasecmp(auth[j], avail_auth_types[i].name, avail_auth_types[i].name_size) == 0) {
+					if (avail_auth_types[i].get_brackets_string)
+						config->auth[x].additional = avail_auth_types[i].get_brackets_string(config, auth[j]+avail_auth_types[i].name_size);
+				
+					config->auth[x].name = talloc_strdup(config, avail_auth_types[i].name);
+					fprintf(stderr, "Enabling '%s' as authentication method\n", avail_auth_types[i].name);
+
+					config->auth[x].amod = avail_auth_types[i].mod;
+					config->auth[x].type |= avail_auth_types[i].type;
+					config->auth[x].enabled = 1;
+					found = 1;
+					x++;
+					if (x >= MAX_AUTH_METHODS) {
+						fprintf(stderr, "You cannot enable more than %d authentication methods\n", x);
+						exit(1);
+					}
+					break;
+				}
+			}
+
+			if (found == 0) {
+				fprintf(stderr, "Unknown or unsupported auth method: %s\n", auth[j]);
+				exit(1);
+			}
+			talloc_free(auth[j]);
+		}
+		config->auth_methods = x;
+
+	}
+	talloc_free(auth);
 }
 
-static void parse_cfg_file(const char* file, struct cfg_st *config, unsigned reload)
+typedef struct acct_types_st {
+	const char *name;
+	unsigned name_size;
+	const struct acct_mod_st *mod;
+	void *(*get_brackets_string)(struct perm_cfg_st *config, const char *);
+} acct_types_st;
+
+static acct_types_st avail_acct_types[] =
+{
+#ifdef HAVE_PAM
+	{NAME("pam"), &pam_acct_funcs, NULL},
+#endif
+#ifdef HAVE_RADIUS
+	{NAME("radius"), &radius_acct_funcs, radius_get_brackets_string},
+#endif
+};
+
+static void figure_acct_funcs(struct perm_cfg_st *config, const char *acct)
+{
+	unsigned i;
+	unsigned found = 0;
+
+	/* Set the accounting method */
+	for (i=0;i<sizeof(avail_acct_types)/sizeof(avail_acct_types[0]);i++) {
+		if (c_strncasecmp(acct, avail_acct_types[i].name, avail_acct_types[i].name_size) == 0) {
+			if (avail_acct_types[i].get_brackets_string)
+				config->acct.additional = avail_acct_types[i].get_brackets_string(config, acct+avail_acct_types[i].name_size);
+
+			if ((avail_acct_types[i].mod->auth_types & config->auth[0].type) == 0) {
+				fprintf(stderr, "You cannot mix the '%s' accounting method with the '%s' authentication method\n", acct, config->auth[0].name);
+				exit(1);
+			}
+
+			config->acct.amod = avail_acct_types[i].mod;
+			config->acct.name = avail_acct_types[i].name;
+			found = 1;
+			break;
+		}
+	}
+
+	if (found == 0) {
+		fprintf(stderr, "Unknown or unsupported accounting method: %s\n", acct);
+		exit(1);
+	}
+	fprintf(stderr, "Setting '%s' as accounting method\n", config->acct.name);
+}
+
+#ifdef HAVE_GSSAPI
+static void parse_kkdcp(struct cfg_st *config, char **urlfw, unsigned urlfw_size)
+{
+	unsigned i, j;
+	char *path, *server, *port, *realm;
+	struct addrinfo hints, *res;
+	int ret;
+	struct kkdcp_st *kkdcp;
+	struct kkdcp_realm_st *kkdcp_realm;
+
+	config->kkdcp = talloc_zero_size(config, urlfw_size*sizeof(kkdcp_st));
+	if (config->kkdcp == NULL) {
+		fprintf(stderr, "memory error\n");
+		exit(1);
+	}
+
+	config->kkdcp_size = 0;
+
+	for (i=0;i<urlfw_size;i++) {
+		path = urlfw[i];
+		realm = strchr(path, ' ');
+		if (realm == NULL) {
+			fprintf(stderr, "Cannot parse kkdcp string: %s\n", path);
+			exit(1);
+		}
+
+		*realm = 0;
+		realm++;
+		while (c_isspace(*realm))
+			realm++;
+
+		server = strchr(realm, ' ');
+		if (server == NULL) {
+			fprintf(stderr, "Cannot parse kkdcp string: %s\n", realm);
+			exit(1);
+		}
+
+		while (c_isspace(*server))
+			server++;
+
+		memset(&hints, 0, sizeof(hints));
+		if (strncmp(server, "udp@", 4) == 0) {
+			hints.ai_socktype = SOCK_DGRAM;
+		} else if (strncmp(server, "tcp@", 4) == 0) {
+			hints.ai_socktype = SOCK_STREAM;
+		} else {
+			fprintf(stderr, "cannot handle protocol %s\n", server);
+			exit(1);
+		}
+		server += 4;
+
+		port = strchr(server, ':');
+		if (port == NULL) {
+			fprintf(stderr, "No server port specified in: %s\n", server);
+			exit(1);
+		}
+		*port = 0;
+		port++;
+
+		ret = getaddrinfo(server, port, &hints, &res);
+		if (ret != 0) {
+			fprintf(stderr, "getaddrinfo(%s) failed: %s\n", server,
+				gai_strerror(ret));
+			exit(1);
+		}
+
+		kkdcp = NULL;
+		/* check if the path is already added */
+		for (j=0;j<config->kkdcp_size;j++) {
+			if (strcmp(path, config->kkdcp[j].url) == 0) {
+				kkdcp = &config->kkdcp[j];
+			}
+		}
+
+		if (kkdcp == NULL) {
+			kkdcp = &config->kkdcp[i];
+			kkdcp->url = talloc_strdup(config->kkdcp, path);
+			config->kkdcp_size++;
+		}
+
+		if (kkdcp->realms_size >= MAX_KRB_REALMS) {
+			fprintf(stderr, "reached maximum number (%d) of realms per URL\n", MAX_KRB_REALMS);
+			exit(1);
+		}
+
+		kkdcp_realm = &kkdcp->realms[kkdcp->realms_size];
+
+		memcpy(&kkdcp_realm->addr, res->ai_addr, res->ai_addrlen);
+		kkdcp_realm->addr_len = res->ai_addrlen;
+		kkdcp_realm->ai_family = res->ai_family;
+		kkdcp_realm->ai_socktype = res->ai_socktype;
+		kkdcp_realm->ai_protocol = res->ai_protocol;
+
+		kkdcp_realm->realm = talloc_strdup(config->kkdcp, realm);
+
+		freeaddrinfo(res);  
+		kkdcp->realms_size++;
+	}
+
+}
+#endif
+
+static void parse_cfg_file(void *pool, const char* file, struct perm_cfg_st *perm_config, unsigned reload)
 {
 tOptionValue const * pov;
 const tOptionValue* val, *prev;
@@ -411,6 +618,11 @@ unsigned prefix = 0, auto_select_group = 0;
 unsigned prefix4 = 0;
 char *tmp;
 unsigned force_cert_auth;
+struct cfg_st *config = perm_config->config;
+#ifdef HAVE_GSSAPI
+char **urlfw = NULL;
+unsigned urlfw_size = 0;
+#endif
 
 	pov = configFileLoad(file);
 	if (pov == NULL && file != NULL && strcmp(file, DEFAULT_CFG_FILE) == 0)
@@ -436,87 +648,70 @@ unsigned force_cert_auth;
 		prev = val;
 	} while((val = optionNextValue(pov, prev)) != NULL);
 
-	config->sup_config_type = SUP_CONFIG_FILE;
+	if (reload == 0) {
+		perm_config->sup_config_type = SUP_CONFIG_FILE;
 
-	READ_MULTI_LINE("auth", auth, auth_size);
-	for (j=0;j<auth_size;j++) {
-		if (c_strncasecmp(auth[j], "pam", 3) == 0) {
-			config->auth_additional = get_brackets_string(config, auth[j]+3);
-			if ((config->auth_types & AUTH_TYPE_USERNAME_PASS) != 0) {
-				fprintf(stderr, "You cannot mix multiple username/password authentication methods\n");
-				exit(1);
-			}
-#ifdef HAVE_PAM
-			amod = &pam_auth_funcs;
-			config->auth_types |= amod->type;
-#else
-			fprintf(stderr, "PAM support is disabled\n");
-			exit(1);
-#endif
-		} else if (strncasecmp(auth[j], "plain", 5) == 0) {
-			if ((config->auth_types & AUTH_TYPE_USERNAME_PASS) != 0) {
-				fprintf(stderr, "You cannot mix multiple username/password authentication methods\n");
-				exit(1);
-			}
+		READ_MULTI_LINE("auth", auth, auth_size);
+		figure_auth_funcs(perm_config, auth, auth_size, 1);
+		auth = NULL;
+		auth_size = 0;
 
-			config->auth_additional = get_brackets_string(config, auth[j]+5);
-			if (config->auth_additional == NULL) {
-				fprintf(stderr, "Format error in %s\n", auth[j]);
-				exit(1);
-			}
-			amod = &plain_auth_funcs;
-			config->auth_types |= amod->type;
-		} else if (strncasecmp(auth[j], "radius", 6) == 0) {
-			if ((config->auth_types & AUTH_TYPE_USERNAME_PASS) != 0) {
-				fprintf(stderr, "You cannot mix multiple username/password authentication methods\n");
-				exit(1);
-			} else {
+		READ_MULTI_LINE("enable-auth", auth, auth_size);
+		figure_auth_funcs(perm_config, auth, auth_size, 0);
+		auth = NULL;
+		auth_size = 0;
 
-#ifdef HAVE_RADIUS
-				const char *p;
-
-				config->auth_additional = get_brackets_string1(config, auth[j]+6);
-				if (config->auth_additional == NULL) {
-					fprintf(stderr, "No configuration specified; error in %s\n", auth[j]);
-					exit(1);
-				}
-
-				p = get_brackets_string2(config, auth[j]+6);
-				if (p != NULL) {
-					if (strcasecmp(p, "groupconfig") != 0) {
-						fprintf(stderr, "No known configuration option: %s\n", p);
-						exit(1);
-					}
-					config->sup_config_type = SUP_CONFIG_RADIUS;
-				}
-				amod = &radius_auth_funcs;
-				config->auth_types |= amod->type;
-#else
-				fprintf(stderr, "Radius support is disabled\n");
-				exit(1);
-#endif
-			}
-		} else if (c_strcasecmp(auth[j], "certificate") == 0) {
-			config->auth_types |= AUTH_TYPE_CERTIFICATE;
-		} else if (c_strcasecmp(auth[j], "certificate[optional]") == 0) {
-			config->auth_types |= AUTH_TYPE_CERTIFICATE_OPT;
-			fprintf(stderr, "The authentication option certificate[optional] is experimental and may be removed in the future\n");
-		} else {
-			fprintf(stderr, "Unknown auth method: %s\n", auth[j]);
+		if (perm_config->auth[0].enabled == 0) {
+			fprintf(stderr, "No authentication method was specified!\n");
 			exit(1);
 		}
-		talloc_free(auth[j]);
+
+		tmp = NULL;
+		READ_STRING("acct", tmp);
+		if (tmp != NULL) {
+			figure_acct_funcs(perm_config, tmp);
+			talloc_free(tmp);
+		}
+
+		PREAD_STRING(pool, "listen-host", perm_config->listen_host);
+		PREAD_STRING(pool, "listen-clear-file", perm_config->unix_conn_file);
+		READ_NUMERIC("tcp-port", perm_config->port);
+		READ_NUMERIC("udp-port", perm_config->udp_port);
+
+		val = get_option("run-as-user", NULL);
+		if (val != NULL && val->valType == OPARG_TYPE_STRING) {
+			const struct passwd* pwd = getpwnam(val->v.strVal);
+			if (pwd == NULL) {
+				fprintf(stderr, "Unknown user: %s\n", val->v.strVal);
+				exit(1);
+			}
+			perm_config->uid = pwd->pw_uid;
+		}
+
+		val = get_option("run-as-group", NULL);
+		if (val != NULL && val->valType == OPARG_TYPE_STRING) {
+			const struct group *grp = getgrnam(val->v.strVal);
+			if (grp == NULL) {
+				fprintf(stderr, "Unknown group: %s\n", val->v.strVal);
+				exit(1);
+			}
+			perm_config->gid = grp->gr_gid;
+		}
 	}
-	talloc_free(auth);
 
 	/* When adding allocated data, remember to modify
 	 * reload_cfg_file();
 	 */
-	READ_STRING("listen-host", config->name);
 	READ_TF("listen-host-is-dyndns", config->is_dyndns, 0);
 
-	READ_NUMERIC("tcp-port", config->port);
-	READ_NUMERIC("udp-port", config->udp_port);
+#ifdef HAVE_GSSAPI
+	READ_MULTI_LINE("kkdcp", urlfw, urlfw_size);
+	if (urlfw_size > 0) {
+		parse_kkdcp(config, urlfw, urlfw_size);
+		talloc_free(urlfw);
+	}
+#endif
+
 	READ_NUMERIC("keepalive", config->keepalive);
 	READ_NUMERIC("dpd", config->dpd);
 	if (config->dpd == 0)
@@ -550,16 +745,11 @@ unsigned force_cert_auth;
 	if (reload == 0 && pid_file[0] == 0)
 		READ_STATIC_STRING("pid-file", pid_file);
 
-	READ_STRING("listen-clear-file", config->unix_conn_file);
-	if (config->unix_conn_file != NULL && (config->auth_types & AUTH_TYPE_CERTIFICATE)) {
-		fprintf(stderr, "The option 'listen-clear-file' cannot be combined with 'auth=certificate'\n");
-		exit(1);
-	}
 
-	READ_STRING("socket-file", config->socket_file_prefix);
-	READ_STRING("occtl-socket-file", config->occtl_socket_file);
-	if (config->occtl_socket_file == NULL)
-		config->occtl_socket_file = talloc_strdup(config, OCCTL_UNIX_SOCKET);
+	PREAD_STRING(perm_config, "socket-file", perm_config->socket_file_prefix);
+	PREAD_STRING(perm_config, "occtl-socket-file", perm_config->occtl_socket_file);
+	if (perm_config->occtl_socket_file == NULL)
+		perm_config->occtl_socket_file = talloc_strdup(perm_config, OCCTL_UNIX_SOCKET);
 
 	val = get_option("session-control", NULL);
 	if (val != NULL) {
@@ -610,7 +800,7 @@ unsigned force_cert_auth;
 	READ_TF("ping-leases", config->ping_leases, 0);
 
 	READ_STRING("tls-priorities", config->priorities);
-	READ_STRING("chroot-dir", config->chroot_dir);
+	PREAD_STRING(perm_config, "chroot-dir", perm_config->chroot_dir);
 
 	READ_NUMERIC("mtu", config->default_mtu);
 
@@ -659,28 +849,25 @@ unsigned force_cert_auth;
 
 	READ_NUMERIC("max-clients", config->max_clients);
 	READ_NUMERIC("min-reauth-time", config->min_reauth_time);
+	config->ban_reset_time = -1;
+	READ_NUMERIC("ban-reset-time", config->ban_reset_time);
+	if (config->ban_reset_time == -1)
+		config->ban_reset_time = DEFAULT_BAN_RESET_TIME;
+
+	config->max_ban_score = -1;
+	READ_NUMERIC("max-ban-score", config->max_ban_score);
+	if (config->max_ban_score == -1)
+		config->max_ban_score = DEFAULT_MAX_BAN_SCORE;
+
+	config->ban_points_wrong_password = DEFAULT_PASSWORD_POINTS;
+	READ_NUMERIC("ban-points-wrong-password", config->ban_points_wrong_password);
+	config->ban_points_connect = DEFAULT_CONNECT_POINTS;
+	READ_NUMERIC("ban-points-connection", config->ban_points_connect);
+	config->ban_points_kkdcp = DEFAULT_KKDCP_POINTS;
+	READ_NUMERIC("ban-points-kkdcp", config->ban_points_kkdcp);
+
 	READ_NUMERIC("max-same-clients", config->max_same_clients);
 
-	val = get_option("run-as-user", NULL);
-	if (val != NULL && val->valType == OPARG_TYPE_STRING) {
-		const struct passwd* pwd = getpwnam(val->v.strVal);
-		if (pwd == NULL) {
-			fprintf(stderr, "Unknown user: %s\n", val->v.strVal);
-			exit(1);
-		}
-		config->uid = pwd->pw_uid;
-	}
-
-	val = get_option("run-as-group", NULL);
-	if (val != NULL && val->valType == OPARG_TYPE_STRING) {
-		const struct group *grp = getgrnam(val->v.strVal);
-		if (grp == NULL) {
-			fprintf(stderr, "Unknown group: %s\n", val->v.strVal);
-			exit(1);
-		}
-		config->gid = grp->gr_gid;
-	}
-
 	READ_STATIC_STRING("device", config->network.name);
 	READ_STRING("cgroup", config->cgroup);
 	READ_STRING("proxy-url", config->proxy_url);
@@ -729,8 +916,9 @@ unsigned force_cert_auth;
 
 	READ_STRING("default-select-group", config->default_select_group);
 	READ_TF("auto-select-group", auto_select_group, 0);
-	if (auto_select_group != 0 && amod != NULL && amod->group_list != NULL) {
-		amod->group_list(config, config->auth_additional, &config->group_list, &config->group_list_size);
+
+	if (auto_select_group != 0 && perm_config->auth[0].amod != NULL && perm_config->auth[0].amod->group_list != NULL) {
+		perm_config->auth[0].amod->group_list(config, perm_config->auth[0].additional, &config->group_list, &config->group_list_size);
 	} else {
 		READ_MULTI_BRACKET_LINE("select-group",
 				config->group_list,
@@ -772,98 +960,114 @@ unsigned force_cert_auth;
 
 
 /* sanity checks on config */
-static void check_cfg(struct cfg_st *config)
+static void check_cfg(struct perm_cfg_st *perm_config)
 {
-	if (config->network.ipv4 == NULL && config->network.ipv6 == NULL) {
+	if (perm_config->config->network.ipv4 == NULL && perm_config->config->network.ipv6 == NULL) {
 		fprintf(stderr, "No ipv4-network or ipv6-network options set.\n");
 		exit(1);
 	}
 
-	if (config->network.ipv4 != NULL && config->network.ipv4_netmask == NULL) {
+	if (perm_config->config->network.ipv4 != NULL && perm_config->config->network.ipv4_netmask == NULL) {
 		fprintf(stderr, "No mask found for IPv4 network.\n");
 		exit(1);
 	}
 
-	if (config->network.ipv6 != NULL && config->network.ipv6_prefix == 0) {
+	if (perm_config->config->network.ipv6 != NULL && perm_config->config->network.ipv6_prefix == 0) {
 		fprintf(stderr, "No prefix found for IPv6 network.\n");
 		exit(1);
 	}
 
-	if (config->banner && strlen(config->banner) > MAX_BANNER_SIZE) {
+	if (perm_config->config->banner && strlen(perm_config->config->banner) > MAX_BANNER_SIZE) {
 		fprintf(stderr, "Banner size is too long\n");
 		exit(1);
 	}
 
-	if (config->cert_size != config->key_size) {
+	if (perm_config->config->cert_size != perm_config->config->key_size) {
 		fprintf(stderr, "The specified number of keys doesn't match the certificates\n");
 		exit(1);
 	}
 
-	if (config->auth_types & AUTH_TYPE_CERTIFICATE) {
-		if (config->cisco_client_compat == 0 && ((config->auth_types & AUTH_TYPE_CERTIFICATE_OPT) != AUTH_TYPE_CERTIFICATE_OPT))
-			config->cert_req = GNUTLS_CERT_REQUIRE;
+	if (perm_config->auth[0].type & AUTH_TYPE_CERTIFICATE) {
+		if (perm_config->config->cisco_client_compat == 0)
+			perm_config->config->cert_req = GNUTLS_CERT_REQUIRE;
 		else
-			config->cert_req = GNUTLS_CERT_REQUEST;
-	}
-
-	if (config->auth_additional != NULL && (config->auth_types & AUTH_TYPE_PLAIN) == AUTH_TYPE_PLAIN) {
-		if (access(config->auth_additional, R_OK) != 0) {
-			fprintf(stderr, "cannot access password file '%s'\n", config->auth_additional);
-			exit(1);
+			perm_config->config->cert_req = GNUTLS_CERT_REQUEST;
+	} else {
+		unsigned i;
+		for (i=1;i<perm_config->auth_methods;i++) {
+			if (perm_config->auth[i].type & AUTH_TYPE_CERTIFICATE) {
+				perm_config->config->cert_req = GNUTLS_CERT_REQUEST;
+				break;
+			}
 		}
 	}
 
-#ifdef ANYCONNECT_CLIENT_COMPAT
-	if (config->cert) {
-		config->cert_hash = calc_sha1_hash(config, config->cert[0], 1);
+	if (perm_config->config->cert_req != 0 && perm_config->config->cert_user_oid == NULL) {
+		fprintf(stderr, "A certificate is requested by the option 'cert-user-oid' is not set\n");
+		exit(1);
 	}
 
-	if (config->xml_config_file) {
-		config->xml_config_hash = calc_sha1_hash(config, config->xml_config_file, 0);
-		if (config->xml_config_hash == NULL && config->chroot_dir != NULL) {
+	if (perm_config->unix_conn_file != NULL && (perm_config->config->cert_req != 0)) {
+		fprintf(stderr, "The option 'listen-clear-file' cannot be combined with 'auth=certificate'\n");
+		exit(1);
+	}
+
+#ifdef ANYCONNECT_CLIENT_COMPAT
+	if (perm_config->config->cert) {
+		perm_config->config->cert_hash = calc_sha1_hash(perm_config->config, perm_config->config->cert[0], 1);
+	}
+
+	if (perm_config->config->xml_config_file) {
+		perm_config->config->xml_config_hash = calc_sha1_hash(perm_config->config, perm_config->config->xml_config_file, 0);
+		if (perm_config->config->xml_config_hash == NULL && perm_config->chroot_dir != NULL) {
 			char path[_POSIX_PATH_MAX];
 
-			snprintf(path, sizeof(path), "%s/%s", config->chroot_dir, config->xml_config_file);
-			config->xml_config_hash = calc_sha1_hash(config, path, 0);
+			snprintf(path, sizeof(path), "%s/%s", perm_config->chroot_dir, perm_config->config->xml_config_file);
+			perm_config->config->xml_config_hash = calc_sha1_hash(perm_config->config, path, 0);
 
-			if (config->xml_config_hash == NULL) {
+			if (perm_config->config->xml_config_hash == NULL) {
 				fprintf(stderr, "Cannot open file '%s'\n", path);
 				exit(1);
 			}
 		}
-		if (config->xml_config_hash == NULL) {
-			fprintf(stderr, "Cannot open file '%s'\n", config->xml_config_file);
+		if (perm_config->config->xml_config_hash == NULL) {
+			fprintf(stderr, "Cannot open file '%s'\n", perm_config->config->xml_config_file);
 			exit(1);
 		}
 	}
 #endif
 
-	if (config->keepalive == 0)
-		config->keepalive = 3600;
+	if (perm_config->config->keepalive == 0)
+		perm_config->config->keepalive = 3600;
 
-	if (config->dpd == 0)
-		config->keepalive = 60;
+	if (perm_config->config->dpd == 0)
+		perm_config->config->dpd = 60;
 
-	if (config->priorities == NULL)
-		config->priorities = "NORMAL:%SERVER_PRECEDENCE:%COMPAT";
+	if (perm_config->config->priorities == NULL)
+		perm_config->config->priorities = talloc_strdup(perm_config->config, "NORMAL:%SERVER_PRECEDENCE:%COMPAT");
 }
 
-int cmd_parser (void *pool, int argc, char **argv, struct cfg_st** config)
+int cmd_parser (void *pool, int argc, char **argv, struct perm_cfg_st** config)
 {
+	*config = talloc_zero(pool, struct perm_cfg_st);
+	if (*config == NULL)
+		exit(1);
 
-	*config = talloc_zero(pool, struct cfg_st);
+	(*config)->config = talloc_zero(*config, struct cfg_st);
+	if ((*config)->config == NULL)
+		exit(1);
 
 	optionProcess( &ocservOptions, argc, argv);
   
 	if (HAVE_OPT(FOREGROUND))
-		(*config)->foreground = 1;
+		(*config)->config->foreground = 1;
 
 	if (HAVE_OPT(PID_FILE)) {
 		strlcpy(pid_file, OPT_ARG(PID_FILE), sizeof(pid_file));
 	}
 
 	if (HAVE_OPT(DEBUG))
-		(*config)->debug = OPT_VALUE_DEBUG;
+		(*config)->config->debug = OPT_VALUE_DEBUG;
 
 	if (HAVE_OPT(CONFIG)) {
 		cfg_file = OPT_ARG(CONFIG);
@@ -872,7 +1076,7 @@ int cmd_parser (void *pool, int argc, char **argv, struct cfg_st** config)
 		exit(1);
 	}
 
-	parse_cfg_file(cfg_file, *config, 0);
+	parse_cfg_file(pool, cfg_file, *config, 0);
 
 	check_cfg(*config);
 
@@ -881,72 +1085,80 @@ int cmd_parser (void *pool, int argc, char **argv, struct cfg_st** config)
 }
 
 #define DEL(x) {talloc_free(x);x=NULL;}
-void clear_cfg_file(struct cfg_st* config)
+void clear_cfg(struct perm_cfg_st* perm_config)
 {
 unsigned i;
 
 #ifdef ANYCONNECT_CLIENT_COMPAT
-	DEL(config->xml_config_file);
-	DEL(config->xml_config_hash);
-	DEL(config->cert_hash);
+	DEL(perm_config->config->xml_config_file);
+	DEL(perm_config->config->xml_config_hash);
+	DEL(perm_config->config->cert_hash);
 #endif
-	DEL(config->cgroup);
-	DEL(config->route_add_cmd);
-	DEL(config->route_del_cmd);
-	DEL(config->per_user_dir);
-	DEL(config->per_group_dir);
-	DEL(config->socket_file_prefix);
-	DEL(config->default_domain);
-	DEL(config->auth_additional);
-	DEL(config->ocsp_response);
-	DEL(config->banner);
-	DEL(config->dh_params_file);
-	DEL(config->name);
-	DEL(config->pin_file);
-	DEL(config->srk_pin_file);
-	DEL(config->ca);
-	DEL(config->crl);
-	DEL(config->cert_user_oid);
-	DEL(config->cert_group_oid);
-	DEL(config->priorities);
-	DEL(config->chroot_dir);
-	DEL(config->connect_script);
-	DEL(config->disconnect_script);
-	DEL(config->proxy_url);
+	DEL(perm_config->config->cgroup);
+	DEL(perm_config->config->route_add_cmd);
+	DEL(perm_config->config->route_del_cmd);
+	DEL(perm_config->config->per_user_dir);
+	DEL(perm_config->config->per_group_dir);
+	DEL(perm_config->config->default_domain);
 
-	DEL(config->network.ipv4);
-	DEL(config->network.ipv4_netmask);
-	DEL(config->network.ipv6);
-	for (i=0;i<config->network.routes_size;i++)
-		DEL(config->network.routes[i]);
-	DEL(config->network.routes);
-	for (i=0;i<config->network.dns_size;i++)
-		DEL(config->network.dns[i]);
-	DEL(config->network.dns);
-	for (i=0;i<config->network.nbns_size;i++)
-		DEL(config->network.nbns[i]);
-	DEL(config->network.nbns);
-	for (i=0;i<config->key_size;i++)
-		DEL(config->key[i]);
-	DEL(config->key);
-	for (i=0;i<config->cert_size;i++)
-		DEL(config->cert[i]);
-	DEL(config->cert);
-	for (i=0;i<config->custom_header_size;i++)
-		DEL(config->custom_header[i]);
-	DEL(config->custom_header);
-	for (i=0;i<config->split_dns_size;i++)
-		DEL(config->split_dns[i]);
-	DEL(config->split_dns);
-	for (i=0;i<config->group_list_size;i++)
-		DEL(config->group_list[i]);
-	DEL(config->group_list);
-	DEL(config->default_select_group);
+	DEL(perm_config->config->ocsp_response);
+	DEL(perm_config->config->banner);
+	DEL(perm_config->config->dh_params_file);
+	DEL(perm_config->config->pin_file);
+	DEL(perm_config->config->srk_pin_file);
+	DEL(perm_config->config->ca);
+	DEL(perm_config->config->crl);
+	DEL(perm_config->config->cert_user_oid);
+	DEL(perm_config->config->cert_group_oid);
+	DEL(perm_config->config->priorities);
+	DEL(perm_config->config->connect_script);
+	DEL(perm_config->config->disconnect_script);
+	DEL(perm_config->config->proxy_url);
+
+#ifdef HAVE_GSSAPI
+	for (i=0;i<perm_config->config->kkdcp_size;i++) {
+		unsigned j;
+		DEL(perm_config->config->kkdcp[i].url);
+		for (j=0;j<perm_config->config->kkdcp[i].realms_size;j++) {
+			DEL(perm_config->config->kkdcp[i].realms[j].realm);
+		}
+	}
+	DEL(perm_config->config->kkdcp);
+#endif
+
+	DEL(perm_config->config->network.ipv4);
+	DEL(perm_config->config->network.ipv4_netmask);
+	DEL(perm_config->config->network.ipv6);
+	for (i=0;i<perm_config->config->network.routes_size;i++)
+		DEL(perm_config->config->network.routes[i]);
+	DEL(perm_config->config->network.routes);
+	for (i=0;i<perm_config->config->network.dns_size;i++)
+		DEL(perm_config->config->network.dns[i]);
+	DEL(perm_config->config->network.dns);
+	for (i=0;i<perm_config->config->network.nbns_size;i++)
+		DEL(perm_config->config->network.nbns[i]);
+	DEL(perm_config->config->network.nbns);
+	for (i=0;i<perm_config->config->key_size;i++)
+		DEL(perm_config->config->key[i]);
+	DEL(perm_config->config->key);
+	for (i=0;i<perm_config->config->cert_size;i++)
+		DEL(perm_config->config->cert[i]);
+	DEL(perm_config->config->cert);
+	for (i=0;i<perm_config->config->custom_header_size;i++)
+		DEL(perm_config->config->custom_header[i]);
+	DEL(perm_config->config->custom_header);
+	for (i=0;i<perm_config->config->split_dns_size;i++)
+		DEL(perm_config->config->split_dns[i]);
+	DEL(perm_config->config->split_dns);
+	for (i=0;i<perm_config->config->group_list_size;i++)
+		DEL(perm_config->config->group_list[i]);
+	DEL(perm_config->config->group_list);
+	DEL(perm_config->config->default_select_group);
 #ifdef HAVE_LIBTALLOC
 	/* our included talloc don't include that */
-	talloc_free_children(config);
+	talloc_free_children(perm_config->config);
 #endif
-	memset(config, 0, sizeof(*config));
+	memset(perm_config->config, 0, sizeof(*perm_config->config));
 
 	return;
 }
@@ -969,6 +1181,9 @@ void print_version(tOptions *opts, tOptDesc *desc)
 #ifdef HAVE_RADIUS
 	fprintf(stderr, "radius, ");
 #endif
+#ifdef HAVE_GSSAPI
+	fprintf(stderr, "gssapi, ");
+#endif
 #ifdef HAVE_PAM
 	fprintf(stderr, "PAM, ");
 #endif
@@ -990,15 +1205,14 @@ void print_version(tOptions *opts, tOptDesc *desc)
 	exit(0);
 }
 
-void reload_cfg_file(void *pool, struct cfg_st* config)
+
+void reload_cfg_file(void *pool, struct perm_cfg_st* perm_config)
 {
-	clear_cfg_file(config);
+	clear_cfg(perm_config);
 
-	memset(config, 0, sizeof(*config));
+	parse_cfg_file(pool, cfg_file, perm_config, 1);
 
-	parse_cfg_file(cfg_file, config, 1);
-
-	check_cfg(config);
+	check_cfg(perm_config);
 
 	return;
 }
