@@ -65,6 +65,9 @@ int send_cookie_auth_reply(main_server_st* s, struct proc_st* proc,
 		msg.session_id.data = proc->dtls_session_id;
 		msg.session_id.len = sizeof(proc->dtls_session_id);
 
+		msg.sid.data = proc->sid;
+		msg.sid.len = sizeof(proc->sid);
+
 		msg.vname = proc->tun_lease.name;
 		msg.user_name = proc->username;
 		msg.group_name = proc->groupname;
@@ -159,9 +162,9 @@ int send_cookie_auth_reply(main_server_st* s, struct proc_st* proc,
 	return 0;
 }
 
-static void apply_default_sup_config(struct cfg_st *config, struct proc_st *proc)
+static void apply_default_sup_config(struct perm_cfg_st *config, struct proc_st *proc)
 {
-	proc->config.deny_roaming = config->deny_roaming;
+	proc->config.deny_roaming = config->config->deny_roaming;
 	proc->config.no_udp = (config->udp_port!=0)?0:1;
 }
 
@@ -201,7 +204,6 @@ struct proc_st *old_proc;
 	proc->dtls_session_id_size = sizeof(proc->dtls_session_id);
 
 	memcpy(proc->sid, cmsg->sid.data, cmsg->sid.len);
-	proc->active_sid = 1;
 
 	/* override the group name in order to load the correct configuration in
 	 * case his group is specified in the certificate */
@@ -211,7 +213,7 @@ struct proc_st *old_proc;
 	/* cookie is good so far, now read config (in order to know
 	 * whether roaming is allowed or not */
 	memset(&proc->config, 0, sizeof(proc->config));
-	apply_default_sup_config(s->config, proc);
+	apply_default_sup_config(s->perm_config, proc);
 
 	/* loads sup config */
 	ret = session_open(s, proc, req->cookie.data, req->cookie.len);
@@ -219,6 +221,8 @@ struct proc_st *old_proc;
 		mslog(s, proc, LOG_INFO, "could not open session");
 		return -1;
 	}
+	/* this hints to call session_close() */
+	proc->active_sid = 1;
 
 	/* Put into right cgroup */
         if (proc->config.cgroup != NULL) {
@@ -259,14 +263,9 @@ struct proc_st *old_proc;
 
 		if (old_proc->pid > 0)
 			kill(old_proc->pid, SIGTERM);
+		mslog(s, proc, LOG_INFO, "resumed cookie session for user '%s' (using %s)", proc->username, proc->tls_ciphersuite);
 	} else {
-		mslog(s, proc, LOG_DEBUG, "new cookie session for (%u)", (unsigned)proc->pid);
-	}
-
-	if (proc->config.require_cert != 0 && cmsg->tls_auth_ok == 0) {
-		mslog(s, proc, LOG_ERR,
-		      "certificate is required for user '%s'", proc->username);
-		return -1;
+		mslog(s, proc, LOG_INFO, "new cookie session for user '%s' (using %s)", proc->username, proc->tls_ciphersuite);
 	}
 
 	if (cmsg->hostname)
