@@ -47,7 +47,8 @@
 
 #define MAX_WAIT_SECS 3
 #define MAX_PIN_SIZE GNUTLS_PKCS11_MAX_PIN_LEN
-#define MAINTAINANCE_TIME 300
+#define MAINTAINANCE_TIME 310
+#define MAX_MSG_SIZE 8*1024
 
 static int need_maintainance = 0;
 static int need_reload = 0;
@@ -188,7 +189,7 @@ static int handle_op(void *pool, int cfd, sec_mod_st * sec, uint8_t type, uint8_
 
 static
 int process_packet(void *pool, int cfd, sec_mod_st * sec, cmd_request_t cmd,
-		   uid_t uid, uint8_t * buffer, size_t buffer_size)
+		   uint8_t * buffer, size_t buffer_size)
 {
 	unsigned i;
 	gnutls_datum_t data, out;
@@ -255,7 +256,6 @@ int process_packet(void *pool, int cfd, sec_mod_st * sec, cmd_request_t cmd,
 			tmsg = cli_stats_msg__unpack(&pa, data.size, data.data);
 			if (tmsg == NULL) {
 				seclog(sec, LOG_ERR, "error unpacking data");
-				ret = ERR_BAD_COMMAND;
 				return -1;
 			}
 
@@ -295,15 +295,48 @@ int process_packet(void *pool, int cfd, sec_mod_st * sec, cmd_request_t cmd,
 			sec_auth_cont_msg__free_unpacked(auth_cont, &pa);
 			return ret;
 		}
+	default:
+		seclog(sec, LOG_WARNING, "unknown type 0x%.2x", cmd);
+		return -1;
+	}
+
+	return 0;
+}
+
+static
+int process_packet_from_main(void *pool, int cfd, sec_mod_st * sec, cmd_request_t cmd,
+		   uint8_t * buffer, size_t buffer_size)
+{
+	gnutls_datum_t data;
+	int ret;
+	PROTOBUF_ALLOCATOR(pa, pool);
+
+	seclog(sec, LOG_DEBUG, "cmd [size=%d] %s\n", (int)buffer_size,
+	       cmd_request_to_str(cmd));
+	data.data = buffer;
+	data.size = buffer_size;
+
+	switch (cmd) {
+	case SM_CMD_AUTH_BAN_IP_REPLY:{
+		BanIpReplyMsg *msg = NULL;
+
+		msg =
+		    ban_ip_reply_msg__unpack(&pa, data.size,
+					     data.data);
+		if (msg == NULL) {
+			seclog(sec, LOG_INFO, "error unpacking auth ban ip reply\n");
+			return -1;
+		}
+
+		handle_sec_auth_ban_ip_reply(cfd, sec, msg);
+		ban_ip_reply_msg__free_unpacked(msg, &pa);
+
+		return 0;
+	}
 	case SM_CMD_AUTH_SESSION_OPEN:
 	case SM_CMD_AUTH_SESSION_CLOSE:{
 			SecAuthSessionMsg *msg;
 
-			if (uid != 0) {
-				seclog(sec, LOG_INFO, "received session open/close from unauthorized uid (%u)\n", (unsigned)uid);
-				return -1;
-			}
-
 			msg =
 			    sec_auth_session_msg__unpack(&pa, data.size,
 						      data.data);
@@ -319,7 +352,7 @@ int process_packet(void *pool, int cfd, sec_mod_st * sec, cmd_request_t cmd,
 		}
 	default:
 		seclog(sec, LOG_WARNING, "unknown type 0x%.2x", cmd);
-		return -1;
+		return ERR_BAD_COMMAND;
 	}
 
 	return 0;
@@ -350,32 +383,28 @@ static void check_other_work(sec_mod_st *sec)
 		}
 
 		sec_mod_client_db_deinit(sec);
-		sec_mod_ban_db_deinit(sec);
 		talloc_free(sec);
 		exit(0);
 	}
 
 	if (need_reload) {
 		seclog(sec, LOG_DEBUG, "reloading configuration");
-		reload_cfg_file(sec, sec->config);
-		sec_auth_reinit(sec, sec->config);
+		reload_cfg_file(sec, sec->perm_config);
 		need_reload = 0;
 	}
 
 	if (need_maintainance) {
 		seclog(sec, LOG_DEBUG, "performing maintenance");
 		cleanup_client_entries(sec);
-		cleanup_banned_entries(sec);
-		seclog(sec, LOG_DEBUG, "active sessions %d, banned entries %d", 
-			sec_mod_client_db_elems(sec),
-			sec_mod_ban_db_elems(sec));
+		seclog(sec, LOG_DEBUG, "active sessions %d", 
+			sec_mod_client_db_elems(sec));
 		alarm(MAINTAINANCE_TIME);
 		need_maintainance = 0;
 	}
 }
 
 static
-void serve_request(sec_mod_st *sec, uid_t uid, int cfd, uint8_t *buffer, unsigned buffer_size)
+int serve_request(sec_mod_st *sec, int cfd, unsigned is_main, uint8_t *buffer, unsigned buffer_size)
 {
 	int ret, e;
 	unsigned cmd, length;
@@ -390,6 +419,7 @@ void serve_request(sec_mod_st *sec, uid_t uid, int cfd, uint8_t *buffer, unsigne
 		e = errno;
 		seclog(sec, LOG_INFO, "error receiving msg head: %s",
 		       strerror(e));
+		ret = -1;
 		goto leave;
 	}
 
@@ -399,6 +429,7 @@ void serve_request(sec_mod_st *sec, uid_t uid, int cfd, uint8_t *buffer, unsigne
 
 	if (length > buffer_size - 4) {
 		seclog(sec, LOG_INFO, "too big message (%d)", length);
+		ret = -1;
 		goto leave;
 	}
 
@@ -408,23 +439,26 @@ void serve_request(sec_mod_st *sec, uid_t uid, int cfd, uint8_t *buffer, unsigne
 		e = errno;
 		seclog(sec, LOG_INFO, "error receiving msg body: %s",
 		       strerror(e));
+		ret = -1;
 		goto leave;
 	}
 
-	ret = process_packet(pool, cfd, sec, cmd, uid, buffer, ret);
+	if (is_main)
+		ret = process_packet_from_main(pool, cfd, sec, cmd, buffer, ret);
+	else
+		ret = process_packet(pool, cfd, sec, cmd, buffer, ret);
 	if (ret < 0) {
 		seclog(sec, LOG_INFO, "error processing data for '%s' command (%d)", cmd_request_to_str(cmd), ret);
 	}
 	
  leave:
- 	talloc_free(pool);
-	close(cfd);
-	return;
+	return ret;
 }
 
 /* sec_mod_server:
  * @config: server configuration
  * @socket_file: the name of the socket
+ * @cmd_fd: socket to exchange commands with main
  *
  * This is the main part of the security module.
  * It creates the unix domain socket identified by @socket_file
@@ -449,12 +483,12 @@ void serve_request(sec_mod_st *sec, uid_t uid, int cfd, uint8_t *buffer, unsigne
  * clients fast without becoming a bottleneck due to private 
  * key operations.
  */
-void sec_mod_server(void *main_pool, struct cfg_st *config, const char *socket_file,
-		    uint8_t cookie_key[COOKIE_KEY_SIZE])
+void sec_mod_server(void *main_pool, struct perm_cfg_st *perm_config, const char *socket_file,
+		    uint8_t cookie_key[COOKIE_KEY_SIZE], int cmd_fd)
 {
 	struct sockaddr_un sa;
 	socklen_t sa_len;
-	int cfd, ret, e;
+	int cfd, ret, e, n;
 	unsigned i, buffer_size;
 	uid_t uid;
 	uint8_t *buffer;
@@ -462,10 +496,18 @@ void sec_mod_server(void *main_pool, struct cfg_st *config, const char *socket_f
 	int sd;
 	sec_mod_st *sec;
 	void *sec_mod_pool;
+	fd_set rd_set;
+	sigset_t emptyset, blockset;
 
 #ifdef DEBUG_LEAKS
 	talloc_enable_leak_report_full();
 #endif
+	sigemptyset(&blockset);
+	sigemptyset(&emptyset);
+	sigaddset(&blockset, SIGALRM);
+	sigaddset(&blockset, SIGTERM);
+	sigaddset(&blockset, SIGINT);
+	sigaddset(&blockset, SIGHUP);
 
 	sec_mod_pool = talloc_init("sec-mod");
 	if (sec_mod_pool == NULL) {
@@ -482,7 +524,8 @@ void sec_mod_server(void *main_pool, struct cfg_st *config, const char *socket_f
 	memcpy(sec->cookie_key, cookie_key, COOKIE_KEY_SIZE);
 	sec->dcookie_key.data = sec->cookie_key;
 	sec->dcookie_key.size = COOKIE_KEY_SIZE;
-	sec->config = talloc_steal(sec, config);
+	sec->perm_config = talloc_steal(sec, perm_config);
+	sec->config = sec->perm_config->config;
 
 	sup_config_init(sec);
 
@@ -501,9 +544,8 @@ void sec_mod_server(void *main_pool, struct cfg_st *config, const char *socket_f
 	ocsignal(SIGTERM, handle_sigterm);
 	ocsignal(SIGALRM, handle_alarm);
 
-	alarm(MAINTAINANCE_TIME);
-
-	sec_auth_init(sec, config);
+	sec_auth_init(sec, perm_config);
+	sec->cmd_fd = cmd_fd;
 
 #ifdef HAVE_PKCS11
 	ret = gnutls_pkcs11_reinit();
@@ -518,10 +560,6 @@ void sec_mod_server(void *main_pool, struct cfg_st *config, const char *socket_f
 		exit(1);
 	}
 
-	if (config->min_reauth_time > 0)
-		sec_mod_ban_db_init(sec);
-
-
 	sd = socket(AF_UNIX, SOCK_STREAM, 0);
 	if (sd == -1) {
 		e = errno;
@@ -539,7 +577,7 @@ void sec_mod_server(void *main_pool, struct cfg_st *config, const char *socket_f
 		exit(1);
 	}
 
-	ret = chown(SOCKET_FILE, config->uid, config->gid);
+	ret = chown(SOCKET_FILE, perm_config->uid, perm_config->gid);
 	if (ret == -1) {
 		e = errno;
 		seclog(sec, LOG_INFO, "could not chown socket '%s': %s", SOCKET_FILE,
@@ -554,15 +592,15 @@ void sec_mod_server(void *main_pool, struct cfg_st *config, const char *socket_f
 		exit(1);
 	}
 
-	ret = load_pins(config, &pins);
+	ret = load_pins(sec->config, &pins);
 	if (ret < 0) {
 		seclog(sec, LOG_ERR, "error loading PIN files");
 		exit(1);
 	}
 
 	/* FIXME: the private key isn't reloaded on reload */
-	sec->key_size = config->key_size;
-	sec->key = talloc_size(sec, sizeof(*sec->key) * config->key_size);
+	sec->key_size = sec->config->key_size;
+	sec->key = talloc_size(sec, sizeof(*sec->key) * sec->config->key_size);
 	if (sec->key == NULL) {
 		seclog(sec, LOG_ERR, "error in memory allocation");
 		exit(1);
@@ -574,19 +612,19 @@ void sec_mod_server(void *main_pool, struct cfg_st *config, const char *socket_f
 		GNUTLS_FATAL_ERR(ret);
 
 		/* load the private key */
-		if (gnutls_url_is_supported(config->key[i]) != 0) {
+		if (gnutls_url_is_supported(sec->config->key[i]) != 0) {
 			gnutls_privkey_set_pin_function(sec->key[i],
 							pin_callback, &pins);
 			ret =
 			    gnutls_privkey_import_url(sec->key[i],
-						      config->key[i], 0);
+						      sec->config->key[i], 0);
 			GNUTLS_FATAL_ERR(ret);
 		} else {
 			gnutls_datum_t data;
-			ret = gnutls_load_file(config->key[i], &data);
+			ret = gnutls_load_file(sec->config->key[i], &data);
 			if (ret < 0) {
 				seclog(sec, LOG_ERR, "error loading file '%s'",
-				       config->key[i]);
+				       sec->config->key[i]);
 				GNUTLS_FATAL_ERR(ret);
 			}
 
@@ -600,38 +638,84 @@ void sec_mod_server(void *main_pool, struct cfg_st *config, const char *socket_f
 		}
 	}
 
+	sigprocmask(SIG_BLOCK, &blockset, &sig_default_set);
+	alarm(MAINTAINANCE_TIME);
 	seclog(sec, LOG_INFO, "sec-mod initialized (socket: %s)", SOCKET_FILE);
 
 	for (;;) {
 		check_other_work(sec);
 
-		sa_len = sizeof(sa);
-		cfd = accept(sd, (struct sockaddr *)&sa, &sa_len);
-		if (cfd == -1) {
-			e = errno;
-			if (e != EINTR) {
-				seclog(sec, LOG_DEBUG,
-				       "sec-mod error accepting connection: %s",
-				       strerror(e));
-			}
+		FD_ZERO(&rd_set);
+		n = 0;
+
+		FD_SET(cmd_fd, &rd_set);
+		n = MAX(n, cmd_fd);
+
+		FD_SET(sd, &rd_set);
+		n = MAX(n, sd);
+
+#ifdef HAVE_PSELECT
+		ret = pselect(n + 1, &rd_set, NULL, NULL, NULL, &emptyset);
+#else
+		sigprocmask(SIG_UNBLOCK, &blockset, NULL);
+		ret = select(n + 1, &rd_set, NULL, NULL, NULL);
+		sigprocmask(SIG_BLOCK, &blockset, NULL);
+#endif
+		if (ret == -1 && errno == EINTR)
 			continue;
+
+		if (ret < 0) {
+			e = errno;
+			seclog(sec, LOG_ERR, "Error in pselect(): %s",
+			       strerror(e));
+			exit(1);
 		}
 
-		/* do not allow unauthorized processes to issue commands
-		 */
-		ret = check_upeer_id("sec-mod", config->debug, cfd, config->uid, config->gid, &uid);
-		if (ret < 0) {
-			seclog(sec, LOG_INFO, "rejected unauthorized connection");
-		} else {
-			buffer_size = 8 * 1024;
+		if (FD_ISSET(cmd_fd, &rd_set)) {
+			buffer_size = MAX_MSG_SIZE;
 			buffer = talloc_size(sec, buffer_size);
 			if (buffer == NULL) {
 				seclog(sec, LOG_ERR, "error in memory allocation");
-				close(cfd);
 			} else {
-				serve_request(sec, uid, cfd, buffer, buffer_size);
+				ret = serve_request(sec, cmd_fd, 1, buffer, buffer_size);
+				if (ret < 0 && ret == ERR_BAD_COMMAND) {
+					seclog(sec, LOG_ERR, "error processing command from main");
+					exit(1);
+				}
+				talloc_free(buffer);
 			}
 		}
+
+		if (FD_ISSET(sd, &rd_set)) {
+			sa_len = sizeof(sa);
+			cfd = accept(sd, (struct sockaddr *)&sa, &sa_len);
+			if (cfd == -1) {
+				e = errno;
+				if (e != EINTR) {
+					seclog(sec, LOG_DEBUG,
+					       "sec-mod error accepting connection: %s",
+					       strerror(e));
+					continue;
+				}
+			}
+
+			/* do not allow unauthorized processes to issue commands
+			 */
+			ret = check_upeer_id("sec-mod", sec->config->debug, cfd, perm_config->uid, perm_config->gid, &uid);
+			if (ret < 0) {
+				seclog(sec, LOG_INFO, "rejected unauthorized connection");
+			} else {
+				buffer_size = MAX_MSG_SIZE;
+				buffer = talloc_size(sec, buffer_size);
+				if (buffer == NULL) {
+					seclog(sec, LOG_ERR, "error in memory allocation");
+				} else {
+					serve_request(sec, cfd, 0, buffer, buffer_size);
+					talloc_free(buffer);
+				}
+			}
+			close(cfd);
+		}
 #ifdef DEBUG_LEAKS
 		talloc_report_full(sec, stderr);
 #endif
