@@ -430,8 +430,8 @@ int open_tun(main_server_st * s, struct proc_st *proc)
 		goto fail;
 	}
 
-	if (s->config->uid != -1) {
-		t = s->config->uid;
+	if (s->perm_config->uid != -1) {
+		t = s->perm_config->uid;
 		ret = ioctl(tunfd, TUNSETOWNER, t);
 		if (ret < 0) {
 			e = errno;
@@ -441,8 +441,8 @@ int open_tun(main_server_st * s, struct proc_st *proc)
 		}
 	}
 #ifdef TUNSETGROUP
-	if (s->config->gid != -1) {
-		t = s->config->uid;
+	if (s->perm_config->gid != -1) {
+		t = s->perm_config->uid;
 		ret = ioctl(tunfd, TUNSETGROUP, t);
 		if (ret < 0) {
 			e = errno;
@@ -499,7 +499,6 @@ int open_tun(main_server_st * s, struct proc_st *proc)
 
 void close_tun(main_server_st * s, struct proc_st *proc)
 {
-	int fd = -1;
 
 	if (proc->tun_lease.fd >= 0) {
 		close(proc->tun_lease.fd);
@@ -507,6 +506,7 @@ void close_tun(main_server_st * s, struct proc_st *proc)
 	}
 
 #ifdef SIOCIFDESTROY
+	int fd = -1;
 	int e, ret;
 	struct ifreq ifr;
 
@@ -525,10 +525,10 @@ void close_tun(main_server_st * s, struct proc_st *proc)
 				proc->tun_lease.name, strerror(e));
 		}
 	}
-#endif
 
 	if (fd != -1)
 		close(fd);
+#endif
 
 	return;
 }
