@@ -41,8 +41,6 @@ fail() {
 }
 
 launch_server() {
-       PARENT=$1;
-       shift;
        $SERV $* >/dev/null 2>&1 &
        LOCALPID="$!";
        trap "[ ! -z \"${LOCALPID}\" ] && kill ${LOCALPID};" 15
@@ -50,15 +48,29 @@ launch_server() {
        LOCALRET="$?"
        if [ "${LOCALRET}" != "0" ] && [ "${LOCALRET}" != "143" ] ; then
                # Houston, we'v got a problem...
-               echo "Failed to launch the server !"
-               test -z "${PARENT}" || kill -10 ${PARENT}
+               exit 1
+       fi
+}
+
+launch_simple_server() {
+       $SERV $* >/dev/null 2>&1 &
+}
+
+launch_debug_server() {
+       valgrind --leak-check=full $SERV $* >out.txt 2>&1 &
+       LOCALPID="$!";
+       trap "[ ! -z \"${LOCALPID}\" ] && kill ${LOCALPID};" 15
+       wait "${LOCALPID}"
+       LOCALRET="$?"
+       if [ "${LOCALRET}" != "0" ] && [ "${LOCALRET}" != "143" ] ; then
+               # Houston, we'v got a problem...
                exit 1
        fi
 }
 
 wait_server() {
 	trap "kill $1" 1 15 2
-	sleep 2
+	sleep 5
 }
 
 trap "fail \"Failed to launch the server, aborting test... \"" 10 
